package ee.yoursit.core.command.bukkit.api;

import lombok.Getter;
import org.bukkit.ChatColor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Stream;

@Getter
public class Argument {
    @NotNull
    private final String name;

    @NotNull
    private final String[] aliases;

    private final boolean consoleAllowed;

    @NotNull
    private final List<Argument> subArguments = new ArrayList<>();

    public Argument() {
        CommandInfo info = getClass().getAnnotation(CommandInfo.class);
        if (info == null)
            throw new IllegalStateException(getClass().getName() + " is not annotated with "
                + CommandInfo.class);
        name = info.name();
        aliases = info.aliases();
        consoleAllowed = info.consoleAllowed();
        create();
    }

    public void create() {
    }

    public void onCommand(@NotNull CommandContext context) throws Exception {
    }

    public Completion onTabComplete(@NotNull CommandContext context) throws Exception {
        return Completion.EMPTY;
    }

    public void onCommandError(@NotNull CommandContext context, @NotNull Throwable error) {
        throw new CommandException("Error whilst handling command execution", error);
    }

    public Completion onTabCompleteError(@NotNull CommandContext context, @NotNull Throwable error) {
        throw new CommandException("Error whilst handling command tab completion", error);
    }

    private void executeCommand(@NotNull CommandContext context) {
        try {
            onCommand(context);
        } catch (Exception e) {
            onCommandError(context, e);
        }
    }

    private Completion executeTabCompletion(@NotNull CommandContext context) {
        try {
            return onTabComplete(context);
        } catch (Exception e) {
            return onTabCompleteError(context, e);
        }
    }

    public final void addArgument(@NotNull Argument argument) {
        subArguments.add(argument);
    }

    public final void execute(@NotNull CommandSender sender, @NotNull String[] args) {
        if (!(sender instanceof Player) && !consoleAllowed) {
            sender.sendMessage(ChatColor.RED + "Console is not allowed to execute this command.");
            return;
        }

        if (args.length == 0 || subArguments.isEmpty()) {
            executeCommand(new CommandContext(sender, args));
            return;
        }

        String name = args[0];
        for (Argument argument : subArguments) {
            if (!argument.test(name))
                continue;

            String[] subArgs = Arrays.copyOfRange(args, 1, args.length);
            argument.execute(sender, subArgs);
            return;
        }

        executeCommand(new CommandContext(sender, args));
    }

    public final Completion complete(@NotNull CommandSender sender, @NotNull String[] args) {
        if (subArguments.isEmpty())
            return executeTabCompletion(new CommandContext(sender, args));

        if (args.length == 0) {
            Stream<String> stream = subArguments
                .stream()
                .map(Argument::getName);
            return Completion.of(stream);
        }

        else if (args.length == 1) {
            String prefix = args[0].toLowerCase();
            Stream<String> stream = subArguments
                .stream()
                .map(Argument::getName)
                .filter(name -> name.startsWith(prefix));
            return Completion.of(stream);
        }

        else {
            String name = args[0].toLowerCase();
            for (Argument argument : subArguments) {
                if (!argument.test(name))
                    continue;

                String[] subArgs = Arrays.copyOfRange(args, 1, args.length);
                return argument.complete(sender, subArgs);
            }
        }

        return executeTabCompletion(new CommandContext(sender, args));
    }

    public final boolean test(@NotNull String name) {
        if (name.equals(this.name))
            return true;

        for (String alias : aliases) {
            if (name.equals(alias))
                return true;
        }

        return false;
    }
}
