package ee.yoursit.core.config;

import ee.yoursit.core.YsmpCore;
import ee.yoursit.core.util.Message;
import lombok.Getter;
import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.configuration.file.YamlConfiguration;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.*;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;

public class Config {
    private final Executor executor = Executors.newSingleThreadExecutor();

    @Getter
    private final String name;
    private final File file;

    @Getter
    private FileConfiguration config;

    public Config(String name) {
        this.name = name;
        file = new File(YsmpCore.INSTANCE.getPlugin().getDataFolder(), name + ".yml");
        reload();
    }

    public void reload() {
        if (!file.exists()) {
            try {
                createConfig();
            } catch (IOException e) {
                Bukkit.getConsoleSender().sendMessage(ChatColor.RED + "Unable to reload config file '" + file.getName() + "'");
                throw new RuntimeException(e);
            }
        }
        load();
    }

    public void reloadAsync() {
        if (!file.exists())
            createConfigAsync(null);
        loadAsync();
    }

    public void reloadAsync(Runnable action) {
        if (!file.exists())
            createConfigAsync(action);
        loadAsync();
    }

    public void save() {
        try {
            config.save(file);
        } catch (IOException e) {
            Bukkit.getConsoleSender().sendMessage(ChatColor.RED + "Unable to save config file '" + file.getName() + "'");
            e.printStackTrace();
        }
    }

    public void saveAsync() {
        executor.execute(this::save);
    }

    public void load() {
        try {
            config = new YamlConfiguration();
            FileInputStream stream = new FileInputStream(file);
            config.load(new InputStreamReader(stream, StandardCharsets.UTF_8));
        } catch (Exception e) {
            throw new IllegalStateException("Unable to load config file", e);
        }
    }

    public void loadAsync() {
        executor.execute(this::load);
    }

    private void createConfig() throws IOException {
        if (!file.createNewFile())
            throw new RuntimeException("Unable to create config.yml");

        InputStream stream = YsmpCore.class.getClassLoader().getResourceAsStream(name + ".yml");
        if (stream == null)
            throw new RuntimeException("File config.yml is missing from class path.");

        BufferedReader reader = new BufferedReader(new InputStreamReader(stream, StandardCharsets.UTF_8));
        StringBuilder content = new StringBuilder();
        String line;

        while ((line = reader.readLine()) != null)
            content.append(line).append('\n');

        try (OutputStreamWriter writer = new OutputStreamWriter(Files.newOutputStream(file.toPath()), StandardCharsets.UTF_8)) {
            writer.write(content.toString());
        }
    }

    private void createConfigAsync(Runnable action) {
        executor.execute(() -> {
            try {
                createConfig();
                action.run();
            } catch (Exception e) {
                Bukkit.getConsoleSender().sendMessage(ChatColor.RED + "Error whilst creating config file '" + file.getName() + "'");
                e.printStackTrace();
            }
        });
    }

    public boolean getOrDefault(String path, boolean defaultValue) {
        if (config.isSet(path))
            return config.getBoolean(path);
        else {
            config.set(path, defaultValue);
            return defaultValue;
        }
    }

    public int getOrDefault(String path, int defaultValue) {
        if (config.isSet(path))
            return config.getInt(path);
        else {
            config.set(path, defaultValue);
            return defaultValue;
        }
    }

    public int getOrDefaultNoSet(String path, int defaultValue) {
        if (config.isSet(path))
            return config.getInt(path);
        else
            return defaultValue;
    }

    public long getOrDefault(String path, long defaultValue) {
        if (config.isSet(path))
            return config.getInt(path);
        else {
            config.set(path, defaultValue);
            return defaultValue;
        }
    }

    public double getOrDefault(String path, double defaultValue) {
        if (config.isSet(path))
            return config.getDouble(path);
        else {
            config.set(path, defaultValue);
            return defaultValue;
        }
    }

    public float getOrDefault(String path, float defaultValue) {
        if (config.isSet(path))
            return  (float) config.getDouble(path);
        else {
            config.set(path, defaultValue);
            return defaultValue;
        }
    }

    public String getOrDefault(String path, String defaultValue) {
        if (config.isSet(path))
            return config.getString(path);
        else {
            if (defaultValue != null)
                config.set(path, defaultValue);
            return defaultValue;
        }
    }

    public String getString(String path) {
        return config.getString(path);
    }

    public long getLong(String path) {
        return config.getLong(path);
    }

    public String getOrDefaultNoSet(String path, String defaultValue) {
        if (config.isSet(path))
            return config.getString(path);
        else
            return defaultValue;
    }

    public String getF(String path) {
        String string = config.getString(path);
        if (string == null)
            throw new IllegalStateException("No such config option: " + path);
        return Message.translate(string);
    }

    public String getOrDefaultF(String path, String defaultValue) {
        String result;
        if (config.isSet(path))
            result = config.getString(path);
        else {
            result = defaultValue;
            if (defaultValue == null)
                return null;
            config.set(path, result);
        }
        return Message.translate(result);
    }

    public String getOrDefaultFLS(String path, String... defaultValue) {
        String result;
        if (config.isSet(path)) {
            StringBuilder builder = new StringBuilder();
            Iterator<String> iterator = config.getStringList(path).iterator();

            while (iterator.hasNext()) {
                builder.append(iterator.next());
                if (iterator.hasNext())
                    builder.append('\n');
            }

            result = builder.toString();
        } else {
            result = String.join("\n", defaultValue);
            config.set(path, Arrays.asList(defaultValue));
        }
        return Message.translate(result);
    }

    public Material getOrDefault(String path, Material defaultValue) {
        if (config.isSet(path))
            return Material.getMaterial(config.getString(path));
        else
            return defaultValue;
    }

    public List<String> getOrDefaultFL(String path, String... defaultValue) {
        if (config.isSet(path))
            return config
                .getStringList(path)
                .stream()
                .map(s -> Message.translate(s))
                .toList();

        else if (defaultValue.length > 0) {
            List<String> result = Arrays
                .stream(defaultValue)
                .map(s -> Message.translate(s))
                .toList();

            config.set(path, result);
            return result;
        }
        return null;
    }

    public List<String> getOrDefault(String path, List<String> defaultValue) {
        if (config.isSet(path))
            return config.getStringList(path);
        else {
            config.set(path, defaultValue);
            return defaultValue;
        }
    }

    public Location getOrDefault(String path, Location defaultValue) {
        if (config.isSet(path)) {
            String world = config.getString(path + ".world");
            if (world == null)
                return null;

            return new Location(
                Bukkit.getWorld(world),
                getOrDefault(path + ".posX", 0D),
                getOrDefault(path + ".posY", 0D),
                getOrDefault(path + ".posZ", 0D),
                getOrDefault(path + ".yaw", 0F),
                getOrDefault(path + ".pitch", 0F)
            );
        } else {
            setLocation(path, defaultValue);
            return defaultValue;
        }
    }

    public Location getLocation(String path) {
        return config.getLocation(path);
    }

    public void setLocation(String path, Location location) {
        if (location == null) {
            config.set(path, new HashMap<>());
            return;
        }

        config.set(path + ".world", location.getWorld().getName());
        config.set(path + ".posX", location.getX());
        config.set(path + ".posY", location.getY());
        config.set(path + ".posZ", location.getZ());
        config.set(path + ".yaw", location.getYaw());
        config.set(path + ".pitch", location.getPitch());
    }

    public Set<String> getKeys(String path) {
        ConfigurationSection section = config.getConfigurationSection(path);
        if (section == null)
            return new HashSet<>();
        return section.getKeys(false);
    }

    public void set(String path, Object value) {
        config.set(path, value);
    }

    public void unset(String path) {
        config.set(path, null);
    }

    public boolean isSet(String path) {
        return config.isSet(path);
    }
}